# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Config reading/loading exceptions
"""


class ConfigException(Exception):  # noqa: N818
    """
    Exception raised in the configuration API
    """
